from requests import HTTPError

ID_MAPPER_DEPENDENCY_NAME = 'IdMapper'


class IdentifierMapperError(Exception):
    """A generic error class when we get an error from Identifier Mapper."""
    dependency_name = ID_MAPPER_DEPENDENCY_NAME

    def __init__(self, http_error: HTTPError = None):
        if http_error:
            self.request = http_error.request
            self.response = http_error.response


class PairNotFoundError(IdentifierMapperError):
    """A Pair is not found in the Identifier Mapper"""


class PartnerNotFoundError(IdentifierMapperError):
    """A Partner could not found in the Identifier Mapper"""



import dataclasses
import random

from mapper.client import IDMapperClient
from mapper.partner import Partner
from mapper.errors import IdentifierMapperError, PartnerNotFoundError
from test.factories.standalone_partner import StandalonePartnerFactory
from test.factories.partner import PartnerFactory
from test.helpers import add_authenticated_payload
import json
import responses
import pytest

from faker import Faker

fake = Faker()


class TestIDMapperPartnerAPI:
    def setup(self):
        self.responses = responses.RequestsMock()
        self.responses.start()
        self.domain = fake.domain_name()
        self.token = fake.sentence()
        secret = {"domain": self.domain, "token": self.token}

        self.target = IDMapperClient(secret)
        self.target.domain = self.domain
        self.target.token = self.token

    def teardown(self):
        self.responses.stop(allow_assert=False)
        self.responses.reset()

    def test_get_partners_by_fields(self) -> None:
        """A partner with full data is successfully retrieved from the Identifier Mapper"""
        fake_partner = PartnerFactory()
        field_choice = random.choice([field.name for field in dataclasses.fields(fake_partner)])
        field_value = dataclasses.asdict(fake_partner)[field_choice]
        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/partners/?{field_choice}={field_value}",
            body=json.dumps([dataclasses.asdict(fake_partner)])
        )

        partners = self.target.get_partners_by_fields(**{field_choice: field_value})
        partner = partners[0]

        assert isinstance(partner, Partner)
        self._assert_identical_partners(fake_partner, partner)

    def test_get_standalone_partner(self) -> None:
        """A standalone partner with data as null is correctly retrieved from the Identifier Mapper"""
        fake_partner = StandalonePartnerFactory()

        field_value = None
        while not field_value:
            field_choice = random.choice([field.name for field in dataclasses.fields(fake_partner)])
            field_value = dataclasses.asdict(fake_partner)[field_choice]

        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/partners/?{field_choice}={field_value}",
            body=json.dumps([dataclasses.asdict(fake_partner)])
        )

        partners = self.target.get_partners_by_fields(**{field_choice: field_value})
        partner = partners[0]

        assert isinstance(partner, Partner)
        self._assert_identical_partners(fake_partner, partner)

    def test_get_partners_by_fields_returns_extra_fields(self) -> None:
        """Additional fields, not in our model, from the partner API do not matter"""
        fake_partner_response = dataclasses.asdict(PartnerFactory())
        fake_partner_response["new_field_" + fake.word()] = fake.sentence()
        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/partners/?foo=bar",
            body=json.dumps([fake_partner_response])
        )

        partner = self.target.get_partners_by_fields(foo='bar')[0]

        assert len(fake_partner_response.keys()) > len(dataclasses.fields(partner))
        assert isinstance(partner, Partner)

    def test_error_is_raised_on_get_partners_by_fields(self):
        """An Identifier Mapper API mapper raises an exception when it receives an error from the service"""
        field = fake.word()
        value = fake.word()
        url = f"https://{self.domain}/api/v1/partners/?{field}={value}"

        self.responses.add(
            responses.GET,
            url,
            status=500
        )

        with pytest.raises(IdentifierMapperError) as exception:
            self.target.get_partners_by_fields(**{field: value})

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 500
        assert exception.value.request.url == url

    def test_not_found_error_is_raised_on_get_partners_by_fields(self):
        """An Identifier Mapper mapper raises a different exception when it receives a 404"""
        field = fake.word()
        value = fake.word()
        url = f"https://{self.domain}/api/v1/partners/?{field}={value}"

        self.responses.add(
            responses.GET,
            url,
            status=404
        )

        with pytest.raises(PartnerNotFoundError) as exception:
            self.target.get_partners_by_fields(**{field: value})

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 404
        assert exception.value.request.url == url

    def test_not_found_error_is_raised_when_get_partners_by_fields_gets_an_empty_list(self) -> None:
        """An Identifier Mapper mapper raises a different exception when it receives a 404"""
        field = fake.word()
        value = fake.word()
        url = f"https://{self.domain}/api/v1/partners/?{field}={value}"

        self.responses.add(
            responses.GET,
            url,
            status=200,
            body="[]"
        )

        with pytest.raises(PartnerNotFoundError) as exception:
            self.target.get_partners_by_fields(**{field: value})

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 200
        assert exception.value.request.url == url

    def test_can_get_a_partner_with_multiple_fields(self):
        """The Identifier Mapper mapper properly forms the query string when it looks up with multiple fields"""
        fake_partner = PartnerFactory()

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/partners/?canvas_domain={fake_partner.canvas_domain}&canvas_account={fake_partner.canvas_account}",
            body=json.dumps([dataclasses.asdict(fake_partner)]),
            auth=f"Token {self.token}"
        )

        partner = self.target.get_partners_by_fields(canvas_domain=fake_partner.canvas_domain,
                                                     canvas_account=fake_partner.canvas_account)[0]
        self._assert_identical_partners(fake_partner, partner)

    def _assert_identical_partners(self, partner1, partner2):
        assert partner1.name == partner2.name
        assert partner1.canvas_domain == partner2.canvas_domain
        assert partner1.canvas_account == partner2.canvas_account
        assert partner1.powerschool_domain == partner2.powerschool_domain
        assert partner1.powerschool_dcid == partner2.powerschool_dcid
        assert partner1.fuji_id == partner2.fuji_id
        assert partner1.canvas_account_uuid == partner2.canvas_account_uuid
        assert partner1.clever_district_id == partner2.clever_district_id
        assert partner1.group_label == partner2.group_label

    def test_get_partner_by_name(self) -> None:
        """A partner with full data is successfully retrieved from the Identifier Mapper by name"""
        fake_partner = PartnerFactory()

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/partners/{fake_partner.name}/",
            body=json.dumps(dataclasses.asdict(fake_partner)),
            auth=f"Token {self.token}"
        )

        partner = self.target.get_partner_by_name(fake_partner.name)

        assert isinstance(partner, Partner)
        self._assert_identical_partners(fake_partner, partner)

    def test_get_partner_by_name_does_not_redirect_without_trailing_slash(self) -> None:
        """A client sends 1 GET to the server and receives a 200 response code
        when the URL in the GET has a trailing slash"""
        fake_partner = PartnerFactory()
        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/partners/{fake_partner.name}",
            body=json.dumps(dataclasses.asdict(fake_partner)),
            status=301
        )

        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/partners/{fake_partner.name}/",
            body=json.dumps(dataclasses.asdict(fake_partner)),
            status=200
        )
        self.target.get_partner_by_id(fake_partner.name)
        assert len(self.responses.calls) == 1
        assert self.responses.calls[0].response.status_code == 200

    def test_get_partner_by_id(self) -> None:
        """A partner with full data is successfully retrieved from the Identifier Mapper by ID"""
        fake_partner = PartnerFactory()

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/partners/{fake_partner.id}/",
            body=json.dumps(dataclasses.asdict(fake_partner)),
            auth=f"Token {self.token}"
        )

        partner = self.target.get_partner_by_id(fake_partner.id)

        assert isinstance(partner, Partner)
        self._assert_identical_partners(fake_partner, partner)

    def test_get_partner_by_id_does_not_redirect_without_trailing_slash(self) -> None:
        """A client sends 1 GET to the server and receives a 200 response code
        when the URL in the GET has a trailing slash"""
        fake_partner = PartnerFactory()
        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/partners/{fake_partner.id}",
            body=json.dumps(dataclasses.asdict(fake_partner)),
            status=301
        )
        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/partners/{fake_partner.id}/",
            body=json.dumps(dataclasses.asdict(fake_partner)),
            status=200
        )
        self.target.get_partner_by_id(fake_partner.id)
        assert len(self.responses.calls) == 1
        assert self.responses.calls[0].response.status_code == 200

    def test_error_is_raised_on_get_partner_by_name(self):
        """An Identifier Mapper API mapper raises an exception when it receives an error from the service"""
        name = fake.word()
        url = f"https://{self.domain}/api/v1/partners/{name}/"

        self.responses.add(
            responses.GET,
            url,
            status=500
        )

        with pytest.raises(IdentifierMapperError) as exception:
            self.target.get_partner_by_name(name)

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 500
        assert exception.value.request.url == url

    def test_not_found_error_is_raised_on_get_partner_by_name(self):
        """An Identifier Mapper mapper raises a different exception when it receives a 404"""
        name = fake.word()
        url = f"https://{self.domain}/api/v1/partners/{name}/"

        self.responses.add(
            responses.GET,
            url,
            status=404
        )

        with pytest.raises(PartnerNotFoundError) as exception:
            self.target.get_partner_by_name(name)

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 404
        assert exception.value.request.url == url

    def test_not_found_error_is_raised_on_get_partner_by_id(self):
        """An Identifier Mapper mapper raises a different exception when it receives a 404"""
        id = fake.uuid4()
        url = f"https://{self.domain}/api/v1/partners/{id}/"

        self.responses.add(
            responses.GET,
            url,
            status=404
        )

        with pytest.raises(PartnerNotFoundError) as exception:
            self.target.get_partner_by_id(id)

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 404
        assert exception.value.request.url == url




from dataclasses import fields
from typing import List, Dict, Any

import requests
from requests import HTTPError

from mapper.errors import IdentifierMapperError, PairNotFoundError, PartnerNotFoundError
from mapper.partner import Partner


class IDMapperClient:
    def __init__(self, secret: Dict):
        self.token = secret['token']
        self.domain = secret['domain']

    def _headers(self):
        headers = {'Authorization': f'Token {self.token}'}
        return headers

    def _get(self, url, params=None, partner_api: bool = False):
        response = requests.get(url, headers=self._headers(), params=params)
        try:
            response.raise_for_status()
        except requests.HTTPError as http_error:
            handle_http_error(http_error, partner_api=partner_api)
        return response

    def _put(self, url, payload):
        response = requests.put(url, headers=self._headers(), json=payload)
        try:
            response.raise_for_status()
        except requests.HTTPError as http_error:
            raise IdentifierMapperError(http_error)
        return response

    def get_pairs(self, service, value):
        url = f"https://{self.domain}/api/v1/pairs/"
        params = {service: value}
        return self._get(url, params).json()

    def get_pair_by_guid(self, guid):
        url = f"https://{self.domain}/api/v1/pairs/{guid}"
        if not guid.startswith('strongmind.guid://'):
            url += '/'

        return self._get(url).json()

    def get_partner_by_name(self, partner_name) -> Partner:
        url = f"https://{self.domain}/api/v1/partners/{partner_name}/"
        responded_partner = self._get(url, partner_api=True).json()
        return self._json_to_partner(responded_partner)

    def get_partner_by_id(self, partner_id: str) -> Partner:
        url = f"https://{self.domain}/api/v1/partners/{partner_id}/"
        responded_partner = self._get(url, partner_api=True).json()
        return self._json_to_partner(responded_partner)

    def get_partners_by_fields(self, **kwargs) -> List[Partner]:
        url = f"https://{self.domain}/api/v1/partners/"
        response = self._get(url, params=kwargs, partner_api=True)
        responded_partners = response.json()
        if not responded_partners:
            raise PartnerNotFoundError(HTTPError(response=response, request=response.request))
        all_partners = []

        for responded_partner in responded_partners:
            all_partners.append(self._json_to_partner(responded_partner))

        return all_partners

    # CAUTION: This should only be used if you know what you're doing
    def put_partner(self, **kwargs):
        partner_name = kwargs.get('name')
        if not partner_name:
            raise PartnerNotFoundError()

        url = f"https://{self.domain}/api/v1/partners/{partner_name}/"
        return self._put(url, kwargs)

    def _json_to_partner(self, partner_json: Dict[str, Any]) -> Partner:
        partner_model_keys = [field.name for field in fields(Partner)]
        shared_keys = set(partner_json) & set(partner_model_keys)

        partner_dict = {}
        for key in shared_keys:
            partner_dict[key] = partner_json[key]
        return Partner(**partner_dict)

    def submit_pairs(self, payload):
        """Send the payload as a bunch of pairs to the Identifier Mapper by
        sending a POST to the Pairs API"""
        url = f"https://{self.domain}/api/v1/pairs/"
        response = requests.post(url, headers=self._headers(), json=payload)

        try:
            response.raise_for_status()
        except requests.HTTPError as http_error:
            raise IdentifierMapperError(http_error)

    def get_or_create_uuid(self, key: str) -> str:
        """Send a GET request to the id mapper uuid route and receive a UUID string"""
        url = f"https://{self.domain}/api/v1/uuid/{key}/"
        return self._get(url).json()


def handle_http_error(http_error, partner_api: bool = False):
    if http_error.response.status_code == 404:
        if partner_api:
            raise PartnerNotFoundError(http_error)
        else:
            raise PairNotFoundError(http_error)
    raise IdentifierMapperError(http_error)



from dataclasses import dataclass, fields


@dataclass
class Partner:
    id: str
    name: str
    canvas_domain: str
    canvas_account: int
    powerschool_domain: str
    powerschool_dcid: int
    powerschool_school_number: int
    fuji_id: int
    default_grade: str
    canvas_account_uuid: str
    clever_district_id: str
    clever_school_id: str
    group_label: str
    roster_source: str
    feature_identity_emails: bool
    canvas_authentication_provider_type: str



from mapper.client import IDMapperClient
from mapper.errors import IdentifierMapperError, PairNotFoundError, PartnerNotFoundError
import json
import responses
import pytest

from faker import Faker

fake = Faker()


class TestIDMapperPostPairsAPI:
    def setup(self):
        self.responses = responses.RequestsMock()
        self.responses.start()

        self.domain = fake.domain_name()
        self.token = fake.sentence()
        self.target = IDMapperClient({"token": self.token, "domain": self.domain})

    def teardown(self):
        self.responses.stop(allow_assert=False)
        self.responses.reset()

    def test_client_can_submit_pairs(self) -> None:
        """An Identifier Mapper mapper can submit a dictionary payload to be treated as pairs"""
        payload = {
            "com.powerschool.user.numbers": {
                "venture": 12345
            },
            "com.instructure.canvas.user.ids": {
                "venture": 58
            }
        }

        self.responses.add(
            responses.POST,
            f"https://{self.target.domain}/api/v1/pairs/",
            status=201
        )

        partners = self.target.submit_pairs(payload)

        assert len(self.responses.calls) == 1
        assert json.loads(self.responses.calls[0].request.body) == payload

    def test_client_cannot_submit_pairs_does_not_redirect_without_trailing_slash(self) -> None:
        """An Identifier Mapper mapper can submit a dictionary payload to be
        treated as pairs when the URL in the POST has a trailing slash"""
        payload = {
            "com.powerschool.user.numbers": {
                "venture": 12345
            },
            "com.instructure.canvas.user.ids": {
                "venture": 58
            }
        }

        self.responses.add(
            responses.POST,
            f"https://{self.target.domain}/api/v1/pairs",
            status=301
        )

        self.responses.add(
            responses.POST,
            f"https://{self.target.domain}/api/v1/pairs/",
            status=201
        )

        partners = self.target.submit_pairs(payload)

        assert len(self.responses.calls) == 1
        assert json.loads(self.responses.calls[0].request.body) == payload

    def test_client_post_failure_raises_exception(self) -> None:
        """An Identifier Mapper mapper will raise an error when the POST fails"""
        url = f"https://{self.target.domain}/api/v1/pairs/"
        self.responses.add(
            responses.POST,
            url,
            status=500
        )

        with pytest.raises(IdentifierMapperError) as exception:
            self.target.submit_pairs({})

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 500
        assert exception.value.request.url == url



from setuptools import setup

setup(
    name='StrongMind-id-mapper-client',
    version='1.3.0',
    packages=['mapper'],
    url='https://github.com/StrongMind/id-mapper-client',
    license='',
    author='team-platform',
    author_email='cinnamon@strongmind.com',
    description='Client for Identifier Mapper'
)



import json

import pytest
import responses
from faker import Faker

from mapper.client import IDMapperClient
from mapper.errors import IdentifierMapperError

fake = Faker()


class TestIDMapperUuidLinkAPI:
    def setup(self):
        self.responses = responses.RequestsMock()
        self.responses.start()
        self.domain = fake.domain_name()
        self.token = fake.sentence()
        secret = {"domain": self.domain, "token": self.token}

        self.target = IDMapperClient(secret)
        self.target.domain = self.domain
        self.target.token = self.token

    def teardown(self):
        self.responses.stop(allow_assert=False)
        self.responses.reset()

    def test_get_uuid_link_by_string_key(self) -> None:
        """A uuid is successfully retrieved from the Identifier Mapper"""
        key = fake.word()
        uuid = fake.uuid4()
        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/uuid/{key}/",
            body=f'"{uuid}"'
        )

        result = self.target.get_or_create_uuid(key=key)

        assert result == uuid

    def test_get_uuid_link_by_string_key_does_not_redirect_without_trailing_slash(self) -> None:
        """A uuid is successfully retrieved from the Identifier Mapper
        when the URL in the GET has a trailing slash"""
        key = fake.word()
        uuid = fake.uuid4()
        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/uuid/{key}",
            body=f'"{uuid}"',
            status=301
        )
        self.responses.add(
            responses.GET,
            f"https://{self.domain}/api/v1/uuid/{key}/",
            body=f'"{uuid}"',
            status=200
        )

        self.target.get_or_create_uuid(key=key)

        assert len(self.responses.calls) == 1
        assert self.responses.calls[0].response.status_code == 200

    def test_get_uuid_link_raises_error(self) -> None:
        """An IdentifierMapperError is raised when the request returns a 4XX or 5XX"""
        key = fake.word()
        uuid = fake.uuid4()
        url = f"https://{self.domain}/api/v1/uuid/{key}/"
        self.responses.add(
            responses.GET,
            url,
            status=500,
        )

        with pytest.raises(IdentifierMapperError) as exception:
            self.target.get_or_create_uuid(key=key)

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 500
        assert exception.value.request.url == url


import json

import pytest
import responses
from faker import Faker

from mapper.client import IDMapperClient
from mapper.errors import IdentifierMapperError, PairNotFoundError, PartnerNotFoundError
from test.factories.partner import PartnerFactory
from test.helpers import add_authenticated_payload
from test.response_factory import response_factory as id_mapper_factory

fake = Faker()


class TestIDMapperClient:
    def setup(self):
        self.responses = responses.RequestsMock()
        self.responses.start()
        self.domain = fake.domain_name()
        self.token = fake.sentence()
        secret = {"domain": self.domain, "token": self.token}

        self.target = IDMapperClient(secret)
        self.target.domain = self.domain
        self.target.token = self.token
        self.partner = PartnerFactory()

    def teardown(self):
        self.responses.stop(allow_assert=False)
        self.responses.reset()

    def test_get_credentials(self):
        """An Identifier Mapper API mapper holds its credentials"""

        assert self.target.token == self.token
        assert self.target.domain == self.domain

    def test_error_is_raised_on_get(self):
        """An Identifier Mapper API mapper raises an error when it receives an error from the service"""
        service = fake.word()
        value = fake.word()
        url = f"https://{self.domain}/api/v1/pairs/?{service}={value}"

        self.responses.add(
            responses.GET,
            url,
            status=500
        )

        with pytest.raises(IdentifierMapperError) as exception:
            self.target.get_pairs(service, value)

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 500
        assert exception.value.request.url == url

    def test_missing_error_on_404(self):
        """An Identifier Mapper API mapper raises an PairNotFound error when it fails to find a pair from the service"""
        service = fake.word()
        value = fake.word()
        url = f"https://{self.domain}/api/v1/pairs/?{service}={value}"

        self.responses.add(
            responses.GET,
            url,
            status=404
        )

        with pytest.raises(PairNotFoundError) as exception:
            self.target.get_pairs(service, value)

        assert exception.value.response is not None
        assert exception.value.request is not None
        assert exception.value.response.status_code == 404
        assert exception.value.request.url == url

    def test_get_returns_api_body(self):
        """An Identifier Mapper API mapper returns the API Body when we try to get a service/value pair from an
        authenticated service """
        service = fake.word()
        value = fake.word()
        subdomain, response_body_callable = id_mapper_factory()
        response_body = response_body_callable()

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/pairs/?{service}={value}",
            body=json.dumps(response_body),
            auth=f"Token {self.token}"
        )

        return_value = self.target.get_pairs(service, value)

        assert return_value == response_body

    def test_get_pairs_with_query_params_does_not_redirect_without_trailing_slash(self):
        """An Identifier Mapper API mapper returns the API Body when we try to get a service/value pair from an
        authenticated service, only when trailing slash is present"""
        service = fake.word()
        value = fake.word()
        subdomain, response_body_callable = id_mapper_factory()
        response_body = response_body_callable()

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/pairs?{service}={value}",
            body=json.dumps(response_body),
            auth=f"Token {self.token}",
            status=301
        )

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/pairs/?{service}={value}",
            body=json.dumps(response_body),
            auth=f"Token {self.token}"
        )

        self.target.get_pairs(service, value)
        assert len(self.responses.calls) == 1
        assert self.responses.calls[0].response.status_code == 200

    def test_client_can_get_pair_by_guid(self):
        """An Identifier Mapper mapper can look up a set of related pairs by a GUID string, nee sourcedId"""
        guid = fake.uuid4()
        subdomain, response_body_callable = id_mapper_factory()
        response_body = response_body_callable()

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/pairs/{guid}/",
            body=json.dumps(response_body),
            auth=f"Token {self.token}"
        )

        return_value = self.target.get_pair_by_guid(guid)

        assert return_value == response_body

    def test_client_can_get_pair_by_strongmind_guid(self):
        """An Identifier Mapper mapper can look up a set of related pairs by a GUID string, nee sourcedId"""
        service = fake.word()
        value = fake.word()
        subdomain, response_body_callable = id_mapper_factory()
        psuedo_guid = f"strongmind.guid://{service}/{subdomain}/{value}"
        response_body = response_body_callable()

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/pairs/{psuedo_guid}",
            body=json.dumps(response_body),
            auth=f"Token {self.token}"
        )

        return_value = self.target.get_pair_by_guid(psuedo_guid)

        assert return_value == response_body

    def test_client_can_put_partner(self):
        """An Identifier Mapper can edit a partner"""
        kwargs = {
            "name": self.partner.name,
            "canvas_authentication_provider_type": fake.random_element(elements=["cas", "openid_connect", ""])
        }

        add_authenticated_payload(
            self.responses,
            responses.PUT,
            f"https://{self.domain}/api/v1/partners/{self.partner.name}/",
            body=json.dumps(kwargs),
            auth=f"Token {self.token}"
        )
        response = self.target.put_partner(**kwargs)
        assert response.status_code == 200

    def test_put_partner_raises_partner_not_found_error(self):
        """An Identifier Mapper raises a PartnerNotFoundError when name is not passed in as a kwarg"""
        kwargs = {
            "canvas_authentication_provider_type": fake.random_element(elements=["cas", "openid_connect", ""])
        }
        with pytest.raises(PartnerNotFoundError):
            self.target.put_partner(**kwargs)

    def test_client_can_get_pair_by_guid_does_not_redirect_without_trailing_slash(self):
        """An Identifier Mapper mapper can look up a set of related pairs by a GUID
        string, only when trailing slash is present"""
        guid = fake.uuid4()
        subdomain, response_body_callable = id_mapper_factory()
        response_body = response_body_callable()

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/pairs/{guid}",
            body=json.dumps(response_body),
            auth=f"Token {self.token}",
            status=301
        )

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/pairs/{guid}/",
            body=json.dumps(response_body),
            auth=f"Token {self.token}",
            status=200
        )

        self.target.get_pair_by_guid(guid)
        assert len(self.responses.calls) == 1
        assert self.responses.calls[0].response.status_code == 200

    def test_client_does_not_add_trailing_slash_for_gets_with_strongmind_guid(self):
        """
        Given we have a StrongMind guid look up
        When the request is handled
        Then the client will be sent the request without a trailing slash
        """
        service = fake.word()
        value = fake.word()
        subdomain, response_body_callable = id_mapper_factory()
        strongmind_guid = f"strongmind.guid://{service}/{subdomain}/{value}"
        response_body = response_body_callable()

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/pairs/{strongmind_guid}",
            body=json.dumps(response_body),
            auth=f"Token {self.token}",
            status=301
        )

        add_authenticated_payload(
            self.responses,
            responses.GET,
            f"https://{self.domain}/api/v1/pairs/{strongmind_guid}/",
            body=json.dumps(response_body),
            auth=f"Token {self.token}",
            status=200
        )

        self.target.get_pair_by_guid(strongmind_guid)
        assert len(self.responses.calls) == 1
        assert self.responses.calls[0].response.status_code == 301

    def test_client_can_submit_pairs(self):
        """An Identifier Mapper mapper can submit pairs"""
        service = fake.word()
        value = fake.word()
        subdomain, response_body_callable = id_mapper_factory()
        request_body = response_body_callable()

        add_authenticated_payload(
            self.responses,
            responses.POST,
            f"https://{self.domain}/api/v1/pairs/",
            body=json.dumps(request_body),
            auth=f"Token {self.token}"
        )

        self.target.submit_pairs(request_body)
